"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.findIssues = void 0;
const node_path_1 = __importDefault(require("node:path"));
const ts_morph_1 = require("ts-morph");
const ts_morph_helpers_1 = require("ts-morph-helpers");
const project_1 = require("./util/project");
const type_1 = require("./util/type");
const dependencies_1 = require("./util/dependencies");
const log_1 = require("./log");
const debug_1 = require("./util/debug");
const lineRewriter = new log_1.LineRewriter();
async function findIssues(configuration) {
    const { workingDir, isShowProgress, report, isDev, jsDocOptions } = configuration;
    const { entryFiles, productionFiles, projectFiles, isIncludeEntryFiles } = configuration;
    const { getUnresolvedDependencies, getUnusedDependencies, getUnusedDevDependencies } = (0, dependencies_1.getDependencyAnalyzer)(configuration);
    const [usedProductionFiles, unreferencedProductionFiles] = (0, project_1.partitionSourceFiles)(projectFiles, productionFiles);
    const [usedEntryFiles, usedNonEntryFiles] = (0, project_1.partitionSourceFiles)(usedProductionFiles, entryFiles);
    (0, debug_1.debugLogSourceFiles)(configuration, 1, 'Used production files', usedProductionFiles);
    (0, debug_1.debugLogSourceFiles)(configuration, 1, 'Unreferenced production files', unreferencedProductionFiles);
    (0, debug_1.debugLogSourceFiles)(configuration, 1, 'Used entry files', usedEntryFiles);
    (0, debug_1.debugLogSourceFiles)(configuration, 1, 'Used non-entry files', usedNonEntryFiles);
    const issues = {
        files: new Set(unreferencedProductionFiles.map(file => file.getFilePath())),
        dependencies: new Set(),
        devDependencies: new Set(),
        unresolved: {},
        exports: {},
        types: {},
        nsExports: {},
        nsTypes: {},
        duplicates: {},
    };
    const counters = {
        files: issues.files.size,
        dependencies: issues.dependencies.size,
        devDependencies: issues.dependencies.size,
        unresolved: 0,
        exports: 0,
        types: 0,
        nsExports: 0,
        nsTypes: 0,
        duplicates: 0,
        processed: issues.files.size,
    };
    const updateProcessingOutput = (item) => {
        if (!isShowProgress)
            return;
        const counter = counters.processed;
        const total = projectFiles.length;
        const percentage = Math.floor((counter / total) * 100);
        const messages = [(0, log_1.getLine)(`${percentage}%`, `of files processed (${counter} of ${total})`)];
        report.files && messages.push((0, log_1.getLine)(unreferencedProductionFiles.length, 'unused files'));
        report.unlisted && messages.push((0, log_1.getLine)(counters.unresolved, 'unlisted dependencies'));
        report.exports && messages.push((0, log_1.getLine)(counters.exports, 'unused exports'));
        report.nsExports && messages.push((0, log_1.getLine)(counters.nsExports, 'unused exports in namespace'));
        report.types && messages.push((0, log_1.getLine)(counters.types, 'unused types'));
        report.nsTypes && messages.push((0, log_1.getLine)(counters.nsTypes, 'unused types in namespace'));
        report.duplicates && messages.push((0, log_1.getLine)(counters.duplicates, 'duplicate exports'));
        if (counter < total) {
            messages.push('');
            messages.push(`Processing: ${node_path_1.default.relative(workingDir, item.filePath)}`);
        }
        lineRewriter.update(messages);
    };
    const addSymbolIssue = (issueType, issue) => {
        const { filePath, symbol } = issue;
        const key = node_path_1.default.relative(workingDir, filePath);
        issues[issueType][key] = issues[issueType][key] ?? {};
        issues[issueType][key][symbol] = issue;
        counters[issueType]++;
        updateProcessingOutput(issue);
    };
    const addProjectIssue = (issueType, issue) => {
        if (!issues[issueType].has(issue.symbol)) {
            issues[issueType].add(issue.symbol);
            counters[issueType]++;
        }
        updateProcessingOutput(issue);
    };
    if (report.dependencies ||
        report.unlisted ||
        report.exports ||
        report.types ||
        report.nsExports ||
        report.nsTypes ||
        report.duplicates) {
        usedProductionFiles.forEach(sourceFile => {
            counters.processed++;
            const filePath = sourceFile.getFilePath();
            if (report.dependencies || report.unlisted) {
                const unresolvedDependencies = getUnresolvedDependencies(sourceFile);
                unresolvedDependencies.forEach(issue => addSymbolIssue('unresolved', issue));
            }
            const exportDeclarations = sourceFile.getExportedDeclarations();
            if (report.duplicates) {
                const duplicateExports = (0, ts_morph_helpers_1.findDuplicateExportedNames)(sourceFile);
                duplicateExports.forEach(symbols => {
                    const symbol = symbols.join('|');
                    addSymbolIssue('duplicates', { filePath, symbol, symbols });
                });
            }
            if (!isIncludeEntryFiles && usedEntryFiles.includes(sourceFile))
                return;
            if (report.exports || report.types || report.nsExports || report.nsTypes) {
                if (!isIncludeEntryFiles) {
                    const uniqueExportedSymbols = new Set([...exportDeclarations.values()].flat());
                    if (uniqueExportedSymbols.size === 1)
                        return;
                }
                exportDeclarations.forEach(declarations => {
                    declarations.forEach(declaration => {
                        const type = (0, type_1.getType)(declaration);
                        if (!report.nsExports && !report.nsTypes) {
                            if (!report.types && type)
                                return;
                            if (!report.exports && !type)
                                return;
                        }
                        if (jsDocOptions.isReadPublicTag && ts_morph_1.ts.getJSDocPublicTag(declaration.compilerNode))
                            return;
                        let identifier;
                        let fakeIdentifier;
                        if (declaration.isKind(ts_morph_1.ts.SyntaxKind.Identifier)) {
                            identifier = declaration;
                        }
                        else if (declaration.isKind(ts_morph_1.ts.SyntaxKind.ArrowFunction) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.ObjectLiteralExpression) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.ArrayLiteralExpression) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.StringLiteral) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.NumericLiteral)) {
                            if (!(0, ts_morph_helpers_1.hasReferencingDefaultImport)(sourceFile)) {
                                fakeIdentifier = 'default';
                            }
                        }
                        else if (declaration.isKind(ts_morph_1.ts.SyntaxKind.FunctionDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.ClassDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.TypeAliasDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.InterfaceDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.EnumDeclaration)) {
                            identifier = declaration.getFirstChildByKindOrThrow(ts_morph_1.ts.SyntaxKind.Identifier);
                        }
                        else if (declaration.isKind(ts_morph_1.ts.SyntaxKind.PropertyAccessExpression)) {
                            identifier = declaration.getLastChildByKindOrThrow(ts_morph_1.ts.SyntaxKind.Identifier);
                        }
                        else {
                            identifier = declaration.getFirstDescendantByKind(ts_morph_1.ts.SyntaxKind.Identifier);
                        }
                        if (identifier || fakeIdentifier) {
                            const identifierText = fakeIdentifier ?? identifier?.getText() ?? '*';
                            if (report.exports && issues.exports[filePath]?.[identifierText])
                                return;
                            if (report.types && issues.types[filePath]?.[identifierText])
                                return;
                            if (report.nsExports && issues.nsExports[filePath]?.[identifierText])
                                return;
                            if (report.nsTypes && issues.nsTypes[filePath]?.[identifierText])
                                return;
                            const refs = identifier?.findReferences() ?? [];
                            if (refs.length === 0) {
                                addSymbolIssue('exports', { filePath, symbol: identifierText });
                            }
                            else {
                                const refFiles = new Set(refs.map(r => r.compilerObject.references.map(r => r.fileName)).flat());
                                const isReferencedOnlyBySelf = refFiles.size === 1 && [...refFiles][0] === filePath;
                                if (!isReferencedOnlyBySelf)
                                    return;
                                if ((0, ts_morph_helpers_1.findReferencingNamespaceNodes)(sourceFile).length > 0) {
                                    if (type) {
                                        addSymbolIssue('nsTypes', { filePath, symbol: identifierText, symbolType: type });
                                    }
                                    else {
                                        addSymbolIssue('nsExports', { filePath, symbol: identifierText });
                                    }
                                }
                                else if (type) {
                                    addSymbolIssue('types', { filePath, symbol: identifierText, symbolType: type });
                                }
                                else {
                                    addSymbolIssue('exports', { filePath, symbol: identifierText });
                                }
                            }
                        }
                    });
                });
            }
        });
    }
    if (report.dependencies) {
        const unusedDependencies = getUnusedDependencies();
        unusedDependencies.forEach(symbol => addProjectIssue('dependencies', { filePath: '', symbol }));
        if (isDev) {
            const unusedDevDependencies = getUnusedDevDependencies();
            unusedDevDependencies.forEach(symbol => addProjectIssue('devDependencies', { filePath: '', symbol }));
        }
    }
    if (isShowProgress)
        lineRewriter.resetLines();
    return { issues, counters };
}
exports.findIssues = findIssues;
